#ifndef CHESSY_MINIMAX_H
#define CHESSY_MINIMAX_H

#include "MoveGeneration.h"

namespace chessy
{
    class Perft;
    class MiniMax;
}


class chessy::Perft
{
    int depth;
    ChessGame& chessGame;
public:
    inline Perft(int depth, ChessGame& cg) :
        depth{ depth }, chessGame{ cg } {}

    size_t search(void);
};


class chessy::MiniMax
{
    ChessGame& game;

    struct BestMove {
        Move move;
        float value;

        inline void overwriteIfBetter(BestMove other)
        {
            if (other.value > value)
                *this = other;
        }

        inline BestMove operator - (void) const
        {
            return { move, -value };
        }
    };

public:
    inline MiniMax(ChessGame& game) :
        game{ game } {}

    Move calculateBest(int depth);

    template<Side side>
    float evaluate(void) const;

private:
    template<Side side>
    BestMove minimax(int depth);
};


#endif // CHESSY_MINIMAX_H
